package com.qmth.distributed.print.upgrade;

import com.alibaba.fastjson.JSON;
import com.qmth.boot.data.upgrade.annotation.DataUpgradeVersion;
import com.qmth.boot.data.upgrade.service.DataUpgradeService;
import com.qmth.boot.data.upgrade.utils.ResourceFileHelper;
import com.qmth.distributed.print.business.entity.ExamCard;
import com.qmth.distributed.print.business.entity.ExamCardModelFour;
import com.qmth.teachcloud.common.bean.dto.mark.PictureConfig;
import com.qmth.teachcloud.common.entity.BasicExam;
import com.qmth.teachcloud.common.entity.MarkQuestion;
import com.qmth.teachcloud.common.entity.SysUser;
import com.qmth.teachcloud.common.enums.ExamModelEnum;
import com.qmth.teachcloud.common.enums.ExceptionResultEnum;
import com.qmth.teachcloud.mark.dto.mark.MarkQuestionAnswerVo;
import com.qmth.teachcloud.mark.dto.mark.manage.MarkerScoreDTO;
import com.qmth.teachcloud.mark.dto.mark.manage.TaskQuestion;
import com.qmth.teachcloud.mark.dto.mark.manage.TrackDTO;
import com.qmth.teachcloud.mark.dto.mark.score.SheetUrlDto;
import com.qmth.teachcloud.mark.dto.mark.score.StudentObjectiveAnswerDto;
import com.qmth.teachcloud.mark.dto.mark.score.StudentPaperDetailDto;
import com.qmth.teachcloud.mark.entity.*;
import com.qmth.teachcloud.mark.enums.CardSource;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DateFormatUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.jdbc.core.BatchPreparedStatementSetter;
import org.springframework.jdbc.core.BeanPropertyRowMapper;
import org.springframework.jdbc.core.JdbcTemplate;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * 升级3.4.4后，3.4.3及以前历史数据的归档
 */
@DataUpgradeVersion("3.4.4")
public class DataUpgrade_3_4_4 implements DataUpgradeService {

    private static final Logger log = LoggerFactory.getLogger(DataUpgrade_3_4_4.class);

    // 一页1000条
    private static int PAGE_SIZE = 500;
    private static String DATE_PATTERN = "yyyy-MM-dd HH:mm:ss";

    @Override
    public void process(JdbcTemplate jdbcTemplate) {
        StringJoiner sj = new StringJoiner(File.separator);
        sj.add(System.getProperty("user.dir")).add("distributed-print").add("install").add("mysql").add("upgrade");
        String upgradePath = sj.toString();
        log.info("数据库脚本根目录：" + upgradePath);
        // 执行升级脚本
//        log.info("开始处理3.4.3历史数据，更新脚本1；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
//        String upgrade_sql1 = upgradePath + File.separator + "3.4.4-upgrade-1.sql";
//        this.execUpgradeSql(jdbcTemplate, upgrade_sql1);
//        log.info("结束处理3.4.3历史数据，更新脚本1；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));

//        log.info("开始处理3.4.3历史数据，数据归档；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
//         执行历史数据归档
        this.execDataArchive(jdbcTemplate);
//        log.info("结束处理3.4.3历史数据，数据归档；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));

        // 归档后中间数据重命名，创建新表
//        log.info("开始处理3.4.4升级脚本2；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
//        String upgrade_sql2 = upgradePath + File.separator + "3.4.4-upgrade-2.sql";
//        this.execUpgradeSql(jdbcTemplate, upgrade_sql2);
//        log.info("结束处理3.4.4升级脚本2；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));

    }

    private void execDataArchive(JdbcTemplate jdbcTemplate) {
        // 用户数据
        Map<Long, SysUser> sysUserMap = this.mapSysUserSql(jdbcTemplate);

        log.info("1.1开始查询待归档课程数据；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
        List<MarkPaper> markPaperList = this.listMarkPaper(jdbcTemplate);
        log.info("1.1结束查询待归档课程数据，共" + markPaperList.size() + "条数据；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
        for (MarkPaper markPaper : markPaperList) {
            Long examId = markPaper.getExamId();
            BasicExam basicExam = getBasicExam(jdbcTemplate, examId);
            String paperNumber = markPaper.getPaperNumber();
            List<MarkStudent> markStudentList = this.listMarkStudent(jdbcTemplate, examId, paperNumber);
            log.info("1.1.1开始归档考试：[" + basicExam.getName() + "]，试卷编号：[" + paperNumber + "]，共" + markStudentList.size() + "条考生数据；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
            List<MarkArchiveStudent> markArchiveStudents = null;
            long count = 0;
            try {
                // 客观题数据
                Map<String, List<MarkQuestionAnswerVo>> objectiveQuestionMap = this.mapObjectiveMarkQuestion(jdbcTemplate, examId, paperNumber);
                // 主观题数据
                List<MarkQuestion> markQuestionList = this.listSubjectiveMarkQuestionSql(jdbcTemplate, examId, paperNumber);

                markArchiveStudents = new ArrayList<>();
                for (MarkStudent markStudent : markStudentList) {
                    MarkArchiveStudent markArchiveStudent = new MarkArchiveStudent();
                    markArchiveStudent.setStudentId(markStudent.getId());
                    markArchiveStudent.setExamId(examId);
                    markArchiveStudent.setPaperNumber(paperNumber);
                    markArchiveStudent.setBasicStudentId(markStudent.getBasicStudentId());
                    markArchiveStudent.setStudentCode(markStudent.getStudentCode());
                    markArchiveStudent.setSheetUrls(buildSheetUrls(jdbcTemplate, markStudent));
                    markArchiveStudent.setSubjectiveQuestions(buildSubjectiveQuestions(jdbcTemplate, markStudent, markQuestionList, sysUserMap));
                    markArchiveStudent.setObjectiveQuestions(buildObjectiveQuestions(objectiveQuestionMap, markStudent));
                    markArchiveStudent.setArchiveCardId(getArchiveCardId(jdbcTemplate, markStudent, basicExam));
                    markArchiveStudent.setCreateTime(System.currentTimeMillis());
                    markArchiveStudents.add(markArchiveStudent);
                }
                count = this.batchInsert(jdbcTemplate, markArchiveStudents);
            } catch (Exception e) {
                // 删除错误数据
                log.info("1.1.1归档失败，删除数据开始。考试：[" + basicExam.getName() + "]，试卷编号：[" + paperNumber + "]，共" + markStudentList.size() + "条考生数据；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
                this.deleteMarkArchiveStudentByExamIdAndPaperNumber(jdbcTemplate, examId, paperNumber);
                log.info("1.1.1归档失败，删除数据结束。考试：[" + basicExam.getName() + "]，试卷编号：[" + paperNumber + "]，共" + markStudentList.size() + "条考生数据；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
                throw new RuntimeException(e.getMessage());
            }
            log.info("1.1.1结束归档考试：[" + basicExam.getName() + "]，试卷编号：[" + paperNumber + "]，共" + markStudentList.size() + "条考生数据；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
            log.info("1.1.2开始更新课程归档状态：[" + basicExam.getName() + "]，试卷编号：[" + paperNumber + "]；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
            if (count == markArchiveStudents.size()) {
                // 更新归档状态
                this.updateMarkPaperArchive(jdbcTemplate, markPaper.getId());
            }
            log.info("1.1.2结束更新课程归档状态：[" + basicExam.getName() + "]，试卷编号：[" + paperNumber + "]；时间：" + DateFormatUtils.format(new Date(), DATE_PATTERN));
        }
    }

    /**
     * 电子卡格式信息
     */
    private String getArchiveCardId(JdbcTemplate jdbcTemplate, MarkStudent markStudent, BasicExam basicExam) {
        String cardContent = getCardContent(jdbcTemplate, markStudent, basicExam);
        if (StringUtils.isBlank(cardContent)) {
            return null;
        }
        String md5 = this.md5String(cardContent);
        int count = this.getMarkArchiveStudentCard(jdbcTemplate, md5);
        if (count == 0) {
            this.insertMarkArchiveStudentCard(jdbcTemplate, md5, cardContent);
        }
        return md5;
    }

    private String getCardContent(JdbcTemplate jdbcTemplate, MarkStudent markStudent, BasicExam basicExam) {
        if (markStudent.getCardNumber() != null) {
            ScanAnswerCard scanAnswerCard = this.getScanAnswerCard(jdbcTemplate, markStudent.getExamId(), markStudent.getCardNumber());
            if (scanAnswerCard != null) {
                if (StringUtils.isNotBlank(scanAnswerCard.getContent())) {
                    return scanAnswerCard.getContent();
                } else {
                    // 自定义卡格式，取电子卡格式的cardId
                    if (ExamModelEnum.MODEL4.equals(basicExam.getExamModel())) {
                        if (scanAnswerCard.getCardId() == null) {
                            List<ScanAnswerCard> scanAnswerCardWebList = this.listScanAnswerCardWeb(jdbcTemplate, scanAnswerCard.getExamId(), scanAnswerCard.getCoursePaperId(), CardSource.WEB);
                            if (CollectionUtils.isNotEmpty(scanAnswerCardWebList)) {
                                ExamCardModelFour examCardModelFour = getExamCardModelFour(jdbcTemplate, scanAnswerCardWebList.get(0).getCardId());
                                return examCardModelFour == null ? null : examCardModelFour.getContent();
                            }
                        } else {
                            ExamCardModelFour examCardModelFour = getExamCardModelFour(jdbcTemplate, scanAnswerCard.getCardId());
                            return examCardModelFour == null ? null : examCardModelFour.getContent();
                        }
                    } else {
                        if (scanAnswerCard.getCardId() == null) {
                            List<ScanAnswerCard> scanAnswerCardWebList = this.listScanAnswerCardWeb(jdbcTemplate, scanAnswerCard.getExamId(), scanAnswerCard.getCoursePaperId(), CardSource.WEB);
                            if (CollectionUtils.isNotEmpty(scanAnswerCardWebList)) {
                                ExamCard examCard = getExamCard(jdbcTemplate, scanAnswerCardWebList.get(0).getCardId());
                                return examCard == null ? null : examCard.getContent();
                            }
                        } else {
                            ExamCard examCard = getExamCard(jdbcTemplate, scanAnswerCard.getCardId());
                            return examCard == null ? null : examCard.getContent();
                        }
                    }
                }
            }
        }
        return null;
    }

    private String md5String(String content) {
        if (StringUtils.isBlank(content)) {
            return null;
        }
        return DigestUtils.md5Hex(content);
    }

    /**
     * 客观题信息
     */
    private String buildObjectiveQuestions(Map<String, List<MarkQuestionAnswerVo>> objectiveQuestionMap, MarkStudent markStudent) {
        if (objectiveQuestionMap.isEmpty()) {
            return null;
        }
        List<String> answers = markStudent.getAnswerList();
        List<String> answerScores = markStudent.getAnswerScoreList();
        List<MarkQuestionAnswerVo> questions = objectiveQuestionMap.get(markStudent.getPaperType());
        int questionCount = questions.size();
        int answerCount = answers.size();
        int answerScoreCount = answerScores.size();

        List<StudentObjectiveAnswerDto> answerDtoList = new ArrayList<>();
        // 已设置客观题
        int maxCount = Math.max(questionCount, answerCount);
        for (int i = 0; i < maxCount; i++) {
            MarkQuestionAnswerVo q = questionCount > i ? questions.get(i) : null;
            String answer = answerCount > i ? answers.get(i) : "#";
            String answerScore = answerScoreCount > i ? answerScores.get(i) : "0";
            StudentObjectiveAnswerDto studentObjectiveAnswerDto = new StudentObjectiveAnswerDto();
            studentObjectiveAnswerDto.setMainNumber(q != null ? q.getMainNumber() : 0);
            studentObjectiveAnswerDto.setSubNumber(q != null ? q.getSubNumber() : 0);
            studentObjectiveAnswerDto.setStandardAnswer(q != null ? q.getAnswer() : null);
            studentObjectiveAnswerDto.setAnswer(answer);
            studentObjectiveAnswerDto.setScore(Double.valueOf(answerScore));
            studentObjectiveAnswerDto.setTotalScore(q != null && q.getTotalScore() != null ? q.getTotalScore() : 0);
            studentObjectiveAnswerDto.setExist(q != null && q.getTotalScore() != null && q.getTotalScore() > 0);
            studentObjectiveAnswerDto.setQuestionType(q != null ? q.getQuestionType() : 0);
            answerDtoList.add(studentObjectiveAnswerDto);
        }
        return JSON.toJSONString(answerDtoList);
    }

    /**
     * 主观题信息
     */
    private String buildSubjectiveQuestions(JdbcTemplate jdbcTemplate, MarkStudent markStudent, List<MarkQuestion> sList, Map<Long, SysUser> sysUserMap) {
        Long examId = markStudent.getExamId();
        String paperNumber = markStudent.getPaperNumber();
        Long studentId = markStudent.getId();
        List<TaskQuestion> taskQuestions = new ArrayList<>();
        List<MarkSubjectiveScore> scoreList = this.listMarkSubjectiveScore(jdbcTemplate, studentId);
        for (int i = 0; i < scoreList.size(); i++) {
            if (sList.size() < i + 1) {
                continue;
            }
            MarkQuestion question = sList.get(i);
            String questionNumber = question.getMainNumber() + "." + question.getSubNumber();
            TaskQuestion step = buildStep(question);
            // 管理员复核时，默认全部为自己评
            step.setSelfMark(true);
            Double score = scoreList.get(i).getScore();
            if (!scoreList.isEmpty() && scoreList.size() == sList.size()) {
                step.setMarkerScore(score);
            }
            // 增加阅卷轨迹列表获取
            List<MarkTask> markTaskList = this.listMarkTask(jdbcTemplate, studentId, question.getGroupNumber());
            // 不管单评还是多评显示所有评卷员给分轨迹
            for (MarkTask markTask : markTaskList) {
                for (MarkTrack markTrack : this.listMarkTrack(jdbcTemplate, markTask.getId(), questionNumber)) {
                    TrackDTO track = new TrackDTO(markTrack);
                    SysUser user = sysUserMap.get(markTask.getUserId());
                    track.setUserId(markTask.getUserId());
                    track.setUserName(user.getRealName() + "(" + user.getLoginName() + ")");
                    step.addMarkTrack(track);
                }

                List<MarkHeaderTrack> markHeaderTracks = this.listMarkHeaderTrack(jdbcTemplate, studentId, questionNumber);
                if (CollectionUtils.isNotEmpty(markHeaderTracks) && markTask.getTaskNumber() == 1) {
                    for (MarkHeaderTrack markHeaderTrack : markHeaderTracks) {
                        TrackDTO track = new TrackDTO(markHeaderTrack);
                        SysUser sysUser = sysUserMap.get(markTask.getHeaderId());
                        if (sysUser != null) {
                            track.setUserId(sysUser.getId());
                            track.setUserName(sysUser.getRealName() + "(" + sysUser.getLoginName() + ")");
                        }
                        track.setHeaderType(markTask.getStatus());
                        step.addHeadTrack(track);
                    }
                }

                // 普通模式无轨迹数据。单独处理评卷员每题得分
                if (CollectionUtils.isEmpty(step.getMarkerTrackList()) && markTask.getMarkerScore() != null) {
                    // 评卷员
                    MarkerScoreDTO markerScoreDTO = new MarkerScoreDTO();
                    markerScoreDTO.setUserId(markTask.getUserId());
                    SysUser markerUser = sysUserMap.get(markTask.getUserId());
                    if (markerUser != null) {
                        markerScoreDTO.setLoginName(markerUser.getLoginName());
                        markerScoreDTO.setUserName(markerUser.getRealName());
                    }
                    markerScoreDTO.setScore(score);
                    markerScoreDTO.setHeader(false);
                    step.addMarkerList(markerScoreDTO);

                    // 科组长
                    if (CollectionUtils.isEmpty(markHeaderTracks) && markTask.getTaskNumber() == 1) {
                        if (!markTask.getMarkerScore().equals(score)) {
                            MarkerScoreDTO headerScoreDTO = new MarkerScoreDTO();
                            headerScoreDTO.setUserId(markTask.getHeaderId());
                            SysUser headerUser = sysUserMap.get(markTask.getHeaderId());
                            if (headerUser != null) {
                                headerScoreDTO.setLoginName(headerUser.getLoginName());
                                headerScoreDTO.setUserName(headerUser.getRealName());
                            }
                            headerScoreDTO.setScore(score);
                            headerScoreDTO.setHeader(true);
                            headerScoreDTO.setHeaderType(markTask.getStatus());
                            step.addMarkerList(headerScoreDTO);
                        }
                    }
                }
            }
            taskQuestions.add(step);
        }
        return JSON.toJSONString(taskQuestions);
    }

    private TaskQuestion buildStep(MarkQuestion question) {
        TaskQuestion step = new TaskQuestion();
        step.setMainNumber(question.getMainNumber());
        step.setSubNumber(question.getSubNumber());
        step.setQuestionType(question.getQuestionType());
        step.setPicList(StringUtils.isNotBlank(question.getPicList()) ? JSON.parseArray(question.getPicList(), PictureConfig.class) : null);
        step.setQuestionId(question.getId());
        step.setTitle(question.getMainTitle());
        step.setDefaultScore(0d);
        step.setMaxScore(question.getTotalScore());
        step.setMinScore(0d);
        step.setIntervalScore(question.getIntervalScore());
        step.setUncalculate(false);
        return step;
    }

    /**
     * 图片信息
     */
    private String buildSheetUrls(JdbcTemplate jdbcTemplate, MarkStudent markStudent) {
        // 原图
        List<SheetUrlDto> sheetUrls = new ArrayList<>();
        List<StudentPaperDetailDto> studentPaperDetailDtoList = this.listData(jdbcTemplate, getSheetUrlsSql(markStudent.getId()), StudentPaperDetailDto.class);
        for (int i = 0; i < studentPaperDetailDtoList.size(); i++) {
            StudentPaperDetailDto studentPaperDetailDto = studentPaperDetailDtoList.get(i);
            sheetUrls.add(new SheetUrlDto(
                    2 * (studentPaperDetailDto.getPaperIndex() - 1) + studentPaperDetailDto.getPageIndex(),
                    studentPaperDetailDto.getSheetPath(),
                    studentPaperDetailDto.getRecogData()));
        }
        return JSON.toJSONString(sheetUrls);
    }

    private void execUpgradeSql(JdbcTemplate jdbcTemplate, String upgradeSql) {
        if (StringUtils.isBlank(upgradeSql)) {
            throw new RuntimeException("没有配置升级sql脚本文件地址");
        }
        File file = new File(upgradeSql);
        if (!file.exists()) {
            throw new RuntimeException("升级sql脚本文件不存在");
        }
        try {
            FileInputStream inputStream = new FileInputStream(upgradeSql);
            String[] sqlList = StringUtils.split(ResourceFileHelper.readContent(inputStream), ";");
            Arrays.stream(sqlList).filter(StringUtils::isNotBlank).forEach(sql -> {
                log.info(sql);
                jdbcTemplate.execute(sql);
            });
        } catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    private <T> T getOne(JdbcTemplate jdbcTemplate, String sql, Class<T> tClass) {
        //将查询的语句封装到List集合中，集合中存储每一个员工实体
        T t = null;
        try {
            t = jdbcTemplate.queryForObject(sql, new BeanPropertyRowMapper<>(tClass));
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception("查询失败：" + e.getMessage());
        }
        return t;
    }

    private <T> T getCount(JdbcTemplate jdbcTemplate, String sql, Class<T> tClass) {
        //将查询的语句封装到List集合中，集合中存储每一个员工实体
        T t = null;
        try {
            t = jdbcTemplate.queryForObject(sql, tClass);
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception("查询失败：" + e.getMessage());
        }
        return t;
    }

    private <T> List<T> listData(JdbcTemplate jdbcTemplate, String sql, Class<T> tClass) {
        //将查询的语句封装到List集合中，集合中存储每一个员工实体
        List<T> list = new ArrayList<>();
        try {
            list = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(tClass));
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception("查询失败：" + e.getMessage());
        }
        return list;
    }

    private int update(JdbcTemplate jdbcTemplate, String sql) {
        try {
            return jdbcTemplate.update(sql);
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception("更新失败：" + e.getMessage());
        }
    }

    private <T> List<T> pageData(JdbcTemplate jdbcTemplate, String sql, Class<T> tClass) {
        if (!sql.contains("limit")) {
            sql += " limit ?,?";
        }
        //将查询的语句封装到List集合中，集合中存储每一个员工实体
        List<T> listAll = new ArrayList<>();
        try {
            List<T> list;
            int pageNumber = 0;
            do {
                int offset = pageNumber * PAGE_SIZE + PAGE_SIZE;
                list = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(tClass), pageNumber, offset);
                if (CollectionUtils.isNotEmpty(list)) {
                    listAll.addAll(list);
                    pageNumber++;
                }
            } while (CollectionUtils.isNotEmpty(list));
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception("查询失败：" + e.getMessage());
        }
        return listAll;
    }

    private <T> List<T> pageData(JdbcTemplate jdbcTemplate, String sql, Class<T> tClass, int pageNumber) {
        //将查询的语句封装到List集合中，集合中存储每一个员工实体
        List<T> listAll = new ArrayList<>();
        try {
            int offset = pageNumber * PAGE_SIZE + PAGE_SIZE;
            List<T> list = jdbcTemplate.query(sql, new BeanPropertyRowMapper<>(tClass), pageNumber, offset);
            if (CollectionUtils.isNotEmpty(list)) {
                listAll.addAll(list);
            }
        } catch (Exception e) {
            throw ExceptionResultEnum.ERROR.exception("查询失败：" + e.getMessage());
        }
        return listAll;
    }

    private String getSheetUrlsSql(Long studentId) {
        StringBuffer sql = new StringBuffer();
        sql.append("SELECT                                                      ");
        sql.append("ssp.student_id studentId,                                   ");
        sql.append("ssp.paper_index paperIndex,                                 ");
        sql.append("ssp.paper_id paperId,                                       ");
        sql.append("sp.card_number cardNumber,                                  ");
        sql.append("spp.page_index pageIndex,                                   ");
        sql.append("spp.sheet_path sheetPath,                                   ");
        sql.append("spp.recog_data recogData                                    ");
        sql.append("FROM                                                        ");
        sql.append("       (SELECT                                              ");
        sql.append("               student_id, paper_index, paper_id            ");
        sql.append("                FROM                                        ");
        sql.append("               scan_student_paper                           ");
        sql.append("                WHERE                                       ");
        sql.append("                student_id = " + studentId + ") ssp         ");
        sql.append("        JOIN                                                ");
        sql.append("scan_paper sp ON ssp.paper_id = sp.id                       ");
        sql.append("AND ssp.paper_index = sp.number                             ");
        sql.append("JOIN                                                        ");
        sql.append("scan_paper_page spp ON sp.id = spp.paper_id                 ");
        sql.append("ORDER BY ssp.paper_index , spp.page_index                   ");
        return sql.toString();
    }

    private Map<String, List<MarkQuestionAnswerVo>> mapObjectiveMarkQuestion(JdbcTemplate jdbcTemplate, Long examId, String paperNumber) {
        StringBuffer sql = new StringBuffer();
        sql.append(" SELECT                                                                     ");
        sql.append("     mqa.paper_type,                                                        ");
        sql.append("     mqa.answer,                                                            ");
        sql.append("     mqa.objective_policy,                                                  ");
        sql.append("     mqa.objective_policy_score,                                            ");
        sql.append("     mq.main_number,                                                        ");
        sql.append("     mq.sub_number,                                                         ");
        sql.append("     mq.total_score,                                                        ");
        sql.append("     mq.question_type                                                       ");
        sql.append(" FROM                                                                       ");
        sql.append("     (SELECT                                                                ");
        sql.append("         *                                                                  ");
        sql.append("     FROM                                                                   ");
        sql.append("         mark_question                                                      ");
        sql.append("     WHERE                                                                  ");
        sql.append("         exam_id = " + examId + " AND paper_number = '" + paperNumber + "'    ");
        sql.append("             AND objective = TRUE) mq                                       ");
        sql.append("         LEFT JOIN                                                          ");
        sql.append("     (SELECT                                                                ");
        sql.append("         *                                                                  ");
        sql.append("     FROM                                                                   ");
        sql.append("         mark_question_answer                                               ");
        sql.append("     WHERE                                                                  ");
        sql.append("         exam_id = " + examId + " AND paper_number = '" + paperNumber + "') mqa    ");
        sql.append("       ON mq.main_number = mqa.main_number                                  ");
        sql.append("         AND mq.sub_number = mqa.sub_number                                 ");
        sql.append(" ORDER BY mq.main_number , mq.sub_number                                    ");

        List<MarkQuestionAnswerVo> questions = this.listData(jdbcTemplate, sql.toString(), MarkQuestionAnswerVo.class);
        Map<String, List<MarkQuestionAnswerVo>> objectiveQuestionMap = questions.stream().collect(Collectors.groupingBy(MarkQuestionAnswerVo::getPaperType));
        return objectiveQuestionMap;
    }

    private List<MarkQuestion> listSubjectiveMarkQuestionSql(JdbcTemplate jdbcTemplate, Long examId, String paperNumber) {
        StringBuffer sql = new StringBuffer();
        sql.append(" SELECT                                                     ");
        sql.append("     *                                                      ");
        sql.append(" FROM                                                       ");
        sql.append("     mark_question mq                                       ");
        sql.append(" WHERE                                                      ");
        sql.append("     mq.exam_id = " + examId + " AND mq.paper_number = '" + paperNumber + "'                 ");
        sql.append("         AND mq.objective = FALSE                           ");
        sql.append(" ORDER BY mq.main_number , mq.sub_number                    ");
        return this.listData(jdbcTemplate, sql.toString(), MarkQuestion.class);
    }

    private List<MarkSubjectiveScore> listMarkSubjectiveScore(JdbcTemplate jdbcTemplate, Long studentId) {
        StringBuffer sql = new StringBuffer();
        sql.append(" SELECT                                                         ");
        sql.append("     *                                                          ");
        sql.append(" FROM                                                           ");
        sql.append("     mark_subjective_score mss                                  ");
        sql.append(" WHERE                                                          ");
        sql.append("     mss.student_id = " + studentId + "                         ");
        sql.append(" ORDER BY mss.main_number , mss.sub_number                      ");
        return this.listData(jdbcTemplate, sql.toString(), MarkSubjectiveScore.class);
    }

    private List<MarkGroup> listMarkGroup(JdbcTemplate jdbcTemplate, Long examId, String paperNumber) {
        String sql = "select * from mark_group where exam_id = " + examId + " and paper_number = '" + paperNumber + "'";
        return this.listData(jdbcTemplate, sql, MarkGroup.class);
    }

    private List<MarkTask> listMarkTask(JdbcTemplate jdbcTemplate, Long studentId, Integer groupNumber) {
        StringBuffer sql = new StringBuffer();
        sql.append(" SELECT                                                             ");
        sql.append("     *                                                              ");
        sql.append(" FROM                                                               ");
        sql.append("     mark_task mt                                                   ");
        sql.append(" WHERE                                                              ");
        sql.append("     mt.student_id = " + studentId + " AND mt.group_number = " + groupNumber + "  ");
        return this.listData(jdbcTemplate, sql.toString(), MarkTask.class);
    }

    private List<MarkTrack> listMarkTrack(JdbcTemplate jdbcTemplate, Long taskId, String questionNumber) {
        String sql = "select * from mark_track where task_id = " + taskId + " and question_number = '" + questionNumber + "'";
        return this.listData(jdbcTemplate, sql, MarkTrack.class);
    }

    private List<MarkHeaderTrack> listMarkHeaderTrack(JdbcTemplate jdbcTemplate, Long studentId, String questionNumber) {
        String sql = "select * from mark_header_track where student_id = " + studentId + " and question_number = '" + questionNumber + "' order by number";
        return this.listData(jdbcTemplate, sql, MarkHeaderTrack.class);
    }

    private Map<Long, SysUser> mapSysUserSql(JdbcTemplate jdbcTemplate) {
        String sql = "select * from sys_user";
        List<SysUser> sysUserList = this.listData(jdbcTemplate, sql, SysUser.class);
        Map<Long, SysUser> sysUserMap = sysUserList.stream().collect(Collectors.toMap(m -> m.getId(), Function.identity()));
        return sysUserMap;
    }

    private List<MarkPaper> listMarkPaper(JdbcTemplate jdbcTemplate) {
        String sql = "select * from mark_paper where status = 'FINISH' and archive = 0";
        return this.listData(jdbcTemplate, sql, MarkPaper.class);
    }

    private List<MarkStudent> listMarkStudent(JdbcTemplate jdbcTemplate, Long examId, String paperNumber) {
        String sql = "select * from mark_student where exam_id = " + examId + " and paper_number = '" + paperNumber + "'";
        return this.listData(jdbcTemplate, sql, MarkStudent.class);
    }

    private ScanAnswerCard getScanAnswerCard(JdbcTemplate jdbcTemplate, Long examId, Integer cardNumber) {
        String sql = "select * from scan_answer_card where exam_id = " + examId + " and number = '" + cardNumber + "'";
        return this.getOne(jdbcTemplate, sql, ScanAnswerCard.class);
    }

    private List<ScanAnswerCard> listScanAnswerCardWeb(JdbcTemplate jdbcTemplate, Long examId, String coursePaperId, CardSource source) {
        String sql = "select * from scan_answer_card where exam_id = " + examId + " and course_paper_id = '" + coursePaperId + "' and source = '" + source + "'";
        return this.listData(jdbcTemplate, sql, ScanAnswerCard.class);
    }

    private BasicExam getBasicExam(JdbcTemplate jdbcTemplate, Long examId) {
        String sql = "select * from basic_exam where id = " + examId;
        return this.getOne(jdbcTemplate, sql, BasicExam.class);
    }

    private ExamCard getExamCard(JdbcTemplate jdbcTemplate, Long cardId) {
        String sql = "select * from exam_card where id = " + cardId;
        return this.getOne(jdbcTemplate, sql, ExamCard.class);
    }

    private ExamCardModelFour getExamCardModelFour(JdbcTemplate jdbcTemplate, Long examId) {
        String sql = "select * from exam_card_model_four where id = " + examId;
        return this.getOne(jdbcTemplate, sql, ExamCardModelFour.class);
    }

    private int getMarkArchiveStudentCard(JdbcTemplate jdbcTemplate, String id) {
        String sql = "select count(1) from mark_archive_student_card where id = '" + id + "'";
        return this.getCount(jdbcTemplate, sql, Integer.class);
    }

    private int updateMarkPaperArchive(JdbcTemplate jdbcTemplate, Long id) {
        String sql = "update mark_paper set archive = 1 where id = " + id;
        return this.update(jdbcTemplate, sql);
    }

    private int insertMarkArchiveStudentCard(JdbcTemplate jdbcTemplate, String id, String cardContent) {
        String sql = "insert into mark_archive_student_card values ('" + id + "','" + cardContent + "')";
        return this.update(jdbcTemplate, sql);
    }

    private int deleteMarkArchiveStudentByExamIdAndPaperNumber(JdbcTemplate jdbcTemplate, Long examId, String paperNumber) {
        String sql = "delete from mark_archive_student where exam_id = " + examId + " and paper_number = '" + paperNumber + "'";
        return this.update(jdbcTemplate, sql);
    }

    private long batchInsert(JdbcTemplate jdbcTemplate, List<MarkArchiveStudent> markArchiveStudents) {
        String sql = "insert into mark_archive_student(student_id, exam_id, paper_number, student_code, basic_student_id, sheet_urls, subjective_questions, objective_questions, archive_card_id, create_time) values (?,?,?,?,?,?,?,?,?,?)";
        int[] ints = jdbcTemplate.batchUpdate(sql, new BatchPreparedStatementSetter() {
            @Override
            public void setValues(PreparedStatement preparedStatement, int i) throws SQLException {
                final MarkArchiveStudent mas = markArchiveStudents.get(i);
                // 参数赋值的时候，序号是从 1 开始的，这一点要注意
                if (mas.getStudentId() != null) {
                    preparedStatement.setLong(1, mas.getStudentId());
                } else {
                    preparedStatement.setNull(1, Types.BIGINT);
                }
                if (mas.getExamId() != null) {
                    preparedStatement.setLong(2, mas.getExamId());
                } else {
                    preparedStatement.setNull(2, Types.BIGINT);
                }
                if (StringUtils.isNotBlank(mas.getPaperNumber())) {
                    preparedStatement.setString(3, mas.getPaperNumber());
                } else {
                    preparedStatement.setNull(3, Types.VARCHAR);
                }
                if (StringUtils.isNotBlank(mas.getStudentCode())) {
                    preparedStatement.setString(4, mas.getStudentCode());
                } else {
                    preparedStatement.setNull(4, Types.VARCHAR);
                }
                if (mas.getBasicStudentId() != null) {
                    preparedStatement.setLong(5, mas.getBasicStudentId());
                } else {
                    preparedStatement.setNull(5, Types.BIGINT);
                }
                if (StringUtils.isNotBlank(mas.getSheetUrls())) {
                    preparedStatement.setString(6, mas.getSheetUrls());
                } else {
                    preparedStatement.setNull(6, Types.VARCHAR);
                }
                if (StringUtils.isNotBlank(mas.getSubjectiveQuestions())) {
                    preparedStatement.setString(7, mas.getSubjectiveQuestions());
                } else {
                    preparedStatement.setNull(7, Types.VARCHAR);
                }
                if (StringUtils.isNotBlank(mas.getObjectiveQuestions())) {
                    preparedStatement.setString(8, mas.getObjectiveQuestions());
                } else {
                    preparedStatement.setNull(8, Types.VARCHAR);
                }
                if (StringUtils.isNotBlank(mas.getArchiveCardId())) {
                    preparedStatement.setString(9, mas.getArchiveCardId());
                } else {
                    preparedStatement.setNull(9, Types.VARCHAR);
                }
                preparedStatement.setLong(10, System.currentTimeMillis());
            }

            @Override
            public int getBatchSize() {
                return markArchiveStudents.size();
            }
        });
        return Arrays.stream(ints).mapToLong(m -> m).sum();
    }
}
